/*
 Copyright (C) 2004 David Bucciarelli (davibu@interfree.it)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.homedns.dade.jcgrid.cmd.povray;

import java.io.*;
import java.text.*;
import java.awt.*;
import java.awt.image.*;
import javax.swing.*;
import javax.imageio.*;

import org.apache.log4j.*;
import org.apache.commons.cli.*;

import org.homedns.dade.jcgrid.*;
import org.homedns.dade.jcgrid.cmd.*;
import org.homedns.dade.jcgrid.client.*;
import org.homedns.dade.jcgrid.message.*;
import org.homedns.dade.jcgrid.worker.impl.povray.*;

public class JCGridClient {
    private final static String className=JCGridClient.class.getName();
    private static Logger log=Logger.getLogger(className);
    private static Logger logDetail=Logger.getLogger("DETAIL."+className);

    //--------------------------------------------------------------------------
    
    public String iniFile="jcgrid.ini";
    public String outputFile="jcgrid-output.png";
    public int width=320;
    public int height=240;
    public int step=64;
    public boolean hasAnimation=false;
    public int startFrame=-1;
    public int endFrame=-1;
    public String outputFormat="png";
    public boolean hasRenderingWindow=true;
    
    //--------------------------------------------------------------------------
    
    public class RenderingFeedback implements POVRenderingClientFeedback {        
        private RenderingWindow rWin;
        private BufferedImage renderingFrameBuffer;
        private Graphics2D renderingFrameBufferG2D;

        private DecimalFormat numberFormat;

        public RenderingFeedback() {            
            rWin=null;
            renderingFrameBuffer=null;
            renderingFrameBufferG2D=null;
            
            numberFormat=new DecimalFormat("00000");
        }

        public void error(String msg, Exception ex) {
            log.warn(msg,ex);
        }

        public void sendingFrgamentRequest(POVWorkRequest req) {
            log.warn("Sending fragment["+req.getFrame()+","+
                    req.getWidth()+"x"+req.getHeight()+","+
                    "("+req.getStartCol()+","+req.getEndCol()+"),"+
                    "("+req.getStartRow()+","+req.getEndRow()+")]");
        }

        public void receivedFragmentResult(POVWorkRequest req, POVWorkResult res,
            int idx, POVRenderingFrame frame) {
            log.warn("Receiving fragment["+req.getFrame()+","+
                req.getWidth()+"x"+req.getHeight()+","+
                "("+req.getStartCol()+","+req.getEndCol()+"),"+
                "("+req.getStartRow()+","+req.getEndRow()+")]");

            if(hasRenderingWindow) {
                renderingFrameBufferG2D.drawImage(frame.getBufferedImage(),null,0,0);
                rWin.update();
            }
        }
        
        public void renderingBegin() {
            // Open rendering window if required

            if(hasRenderingWindow) {
                // Creating rendering framebuffer

                renderingFrameBuffer=new BufferedImage(
                    width,height,BufferedImage.TYPE_3BYTE_BGR);
                renderingFrameBufferG2D=renderingFrameBuffer.createGraphics();

                rWin=new RenderingWindow(renderingFrameBuffer);
                rWin.setVisible(true);
                rWin.update();
            } else
                rWin=null;
        }
        
        public void renderingEnd() {
            if(hasRenderingWindow)
                    rWin.dispose();
        }
        
        public void setProgressMaximum(int max) {
        }
        
        public void setProgressMinimum(int min) {
        }
        
        public void setProgressValue(int val) {
        }
        
        public void setRenderingTime(POVWorkRequest req, long dt) {
            double totTime=dt/(60.0*1000.0);
            if(dt>0.0001) {
                long pixSec;
                if(hasAnimation)
                    pixSec=(1000*(endFrame-startFrame+1)*req.getWidth()*req.getHeight())/dt;
                else
                    pixSec=(1000*req.getWidth()*req.getHeight())/dt;

                log.warn("Total rendering time: "+totTime+"min ("+
                    pixSec+"pixel/sec)");
            } else
                log.warn("Total rendering time: N/A");
        }

        public void completeFrame(int idx, POVRenderingFrame frame) {
            try {
                String fileName;
                if(hasAnimation)
                    fileName=numberFormat.format(idx)+"-"+outputFile;
                else                
                    fileName=outputFile;

                log.warn("Complete frame: "+fileName+" ["+idx+"]");
                
                File f=new File(fileName);
                ImageIO.write(frame.getBufferedImage(),outputFormat,f);
            } catch(Exception ex) {
                log.warn("Error while saving file: "+outputFile,ex);
                System.exit(0);
            }
        }
    }
    
    public void doRendering(GridNodeClientConfig config,
        POVWorkRequest povwr) throws Exception  {
        RenderingFeedback rfback=new RenderingFeedback();
        POVRenderingClient rc;
        if(hasAnimation)
            rc=new POVRenderingClient(
                config,rfback,povwr,step,
                startFrame,endFrame);
        else
            rc=new POVRenderingClient(
                config,rfback,povwr,step);

        rc.start();
        rc.join();
    }
    
    //--------------------------------------------------------------------------
    
    public static void main(String[] args) {
        try {
            // Setup log4j

            MainCmd.setUpLog4J("client",true);

            // Setup GridServer

            log.warn("-----------------------------------------------");
            log.warn("-- JCGridClient POVRay v"+Version.RELEASE);
            log.warn("-----------------------------------------------");

            JCGridClient client=new JCGridClient();
            GridNodeClientConfig config=new GridNodeClientConfig();

            // Other command line options

            Options options=new Options();
            options.addOption("f",true,"set the povray ini file");
            options.addOption("o",true,"set the povray output file");
            options.addOption("t",true,"set the povray output file format [i.e. png, ppm, jpg, etc.]");
            options.addOption("x",true,"set the povray output file width");
            options.addOption("y",true,"set the povray output file height");
            options.addOption("g",true,"set the povray image fragment size");
            options.addOption("r",false,"disable rendering window");
            options.addOption("b",true,"set the povray animation start frame");
            options.addOption("h",true,"set the povray animation end frame");
            
            try {
                CommandLine cmd=MainCmd.parseCommonOptions(
                    options,config,args);
                
                // POVRay options

                if(cmd.hasOption("f"))
                    client.iniFile=cmd.getOptionValue("f");
                if(cmd.hasOption("o"))
                    client.outputFile=cmd.getOptionValue("o");
                if(cmd.hasOption("t"))
                    client.outputFormat=cmd.getOptionValue("t");
                if(cmd.hasOption("x"))
                    client.width=Integer.parseInt(cmd.getOptionValue("x"));
                if(cmd.hasOption("y"))
                    client.height=Integer.parseInt(cmd.getOptionValue("y"));
                if(cmd.hasOption("g"))
                    client.step=Integer.parseInt(cmd.getOptionValue("g"));
                if(cmd.hasOption("r"))
                    client.hasRenderingWindow=false;
                if(cmd.hasOption("b"))
                    client.startFrame=Integer.parseInt(cmd.getOptionValue("b"));
                if(cmd.hasOption("h"))
                    client.endFrame=Integer.parseInt(cmd.getOptionValue("h"));

                if(cmd.getArgs().length>0)
                    throw new Exception("Unknown command line option");
                if(((client.startFrame==-1) && (client.endFrame!=-1)) ||
                    ((client.endFrame==-1) && (client.startFrame!=-1)))
                    throw new Exception("Both start and end animation frame must be set");
                    
                if((client.startFrame!=-1) && (client.endFrame!=-1)) {
                    client.hasAnimation=true;

                    if(client.endFrame<client.startFrame)
                        throw new Exception("End animation frame must be greater than end animation frame");
                }
            } catch(Exception ex) {
                log.warn("Error while parsing command line",ex);

                HelpFormatter formatter=new HelpFormatter();
                formatter.printHelp("JCGridClient",options);

                System.exit(0);
            }


            // Creating work requests
            
            POVWorkRequest povwr=new POVWorkRequest(
                config.getSessionName(),0,
                client.iniFile,client.width,
                client.height,0,client.width-1,0,client.height-1);
            povwr.setAnimationEnabled(client.hasAnimation);
            
            client.doRendering(config,povwr);

            // Rendering done

            log.warn("Done.");
        } catch(Exception ex) {
            log.warn("Error",ex);
            System.exit(0);
        }
    }
}
