/*
 Copyright (C) 2004 David Bucciarelli (davibu@interfree.it)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.homedns.dade.jcgrid.server;

import java.io.*;
import java.net.*;
import java.util.*;
import java.security.*;

import org.apache.log4j.*;

import org.homedns.dade.jcgrid.*;
import org.homedns.dade.jcgrid.vfs.*;
import org.homedns.dade.jcgrid.message.*;
import org.homedns.dade.jcgrid.util.*;

public class GridServer extends GridNode {
    private final static String className=GridServer.class.getName();
    private static Logger log=Logger.getLogger(className);
    private static Logger logDetail=Logger.getLogger("DETAIL."+className);

    private ClientMainDemon clientMainDemon;
    private WorkerMainDemon workerMainDemon;
    private AdminMainDemon adminMainDemon;

    private HandlerThreads connectedClients;
    private HandlerThreads connectedWorkers;
    private HandlerThreads connectedAdmins;

    private WorkRequestQueue workRequestQueue;
    
    private vfsSessionPool sessions;

    public GridServer() {
        super(new GridNodeServerConfig());

        if(log.isDebugEnabled()) log.debug("Start GridServer()");

        clientMainDemon=null;
        workerMainDemon=null;
        adminMainDemon=null;

        connectedClients=new HandlerThreads();
        connectedWorkers=new HandlerThreads();
        connectedAdmins=new HandlerThreads();

        workRequestQueue=new WorkRequestQueue();

        sessions=null;

        if(log.isDebugEnabled()) log.debug("End GridServer()");
    }

    //----------------------- Work requests & results --------------------------

    protected void pushWorkRequest(WorkRequest wr) {
        if(log.isDebugEnabled()) log.debug("Start pushWorkRequest()");

        workRequestQueue.push(wr);

        if(log.isDebugEnabled()) log.debug("End pushWorkRequest()");
    }

    protected WorkRequest popWorkRequest(long timeout) throws InterruptedException {
        if(log.isDebugEnabled()) log.debug("Start popWorkRequest("+timeout+")");

        WorkRequest wr=(WorkRequest)workRequestQueue.pop(timeout);

        if(log.isDebugEnabled()) log.debug("End popWorkRequest("+wr+")");

        return wr;
    }

    protected WorkRequest popWorkRequest() throws InterruptedException {
        if(log.isDebugEnabled()) log.debug("Start popWorkRequest()");

        WorkRequest wr=(WorkRequest)workRequestQueue.pop();

        if(log.isDebugEnabled()) log.debug("End popWorkRequest("+wr+")");

        return wr;
    }

    protected void removePendingWorkRequests(String sessionName) {
        if(log.isDebugEnabled()) log.debug("Start removePendingWorkRequests("+sessionName+")");

        workRequestQueue.removePendingWorkRequests(sessionName);

        if(log.isDebugEnabled()) log.debug("End removePendingWorkRequests()");
    }

    protected void pushWorkResult(WorkResult wr) throws Exception {
        if(log.isDebugEnabled()) log.debug("Start pushWorkResult("+wr+")");

        String sessionName=wr.getSessionName();

        ClientHandlerThread cht=(ClientHandlerThread)connectedClients.
            getConnectedHandler(sessionName);

        // If the session doesn't exist do nothing

        if(cht!=null)
            cht.pushWorkResult(wr);

        if(log.isDebugEnabled()) log.debug("End pushWorkResult()");
    }

    //----------------------------- Start & Stop -------------------------------

    public void start() throws Exception {
        if(log.isDebugEnabled()) log.debug("Start start()");

        super.start();

        if(this.getNodeConfig().getGridConfig().getUseVFS()) {
            long maxCacheSize=((GridNodeServerConfig)this.getNodeConfig()).getMaxCacheSize();
            sessions=new vfsSessionPool(
                this.getNodeConfig().getWorkingDir(),maxCacheSize);
            sessions.start();
            log.warn("VFS Session cache size: "+sessions.getSize()/(1024*1024)+"/"+
                maxCacheSize/(1024*1024)+"MB");
        } else
            sessions=null;

        clientMainDemon=new ClientMainDemon(this);
        clientMainDemon.start();

        workerMainDemon=new WorkerMainDemon(this);
        workerMainDemon.start();

        adminMainDemon=new AdminMainDemon(this);
        adminMainDemon.start();

        log.warn("Ok");
        
        if(log.isDebugEnabled()) log.debug("End start()");
    }

    public void stop() throws InterruptedException {
        if(log.isDebugEnabled()) log.debug("Start stop()");

        clientMainDemon.interrupt();
        clientMainDemon.join();

        workerMainDemon.interrupt();
        workerMainDemon.join();

        adminMainDemon.interrupt();
        adminMainDemon.join();

        if(log.isDebugEnabled()) log.debug("End stop()");
    }

    //------------------------------- Handlers ---------------------------------

    protected HandlerThreads getClientHandlers() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getClientHandlers()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getClientHandlers("+connectedClients+")");
        
        return connectedClients;
    }

    protected HandlerThreads getWorkerHandlers() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getWorkerHandlers()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getWorkerHandlers("+connectedWorkers+")");
        
        return connectedWorkers;
    }

    protected HandlerThreads getAdminHandlers() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getAdminHandlers()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getAdminHandlers("+connectedAdmins+")");
        
        return connectedAdmins;
    }

    //------------------------------- Get & Set --------------------------------

    public vfsSessionPool getVFSSessionPool() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getVFSSessionPool()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getVFSSessionPool("+sessions+")");
        
        return sessions;
    }

    public int getRequestQueueSize() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getRequestQueueSize()");

        int res=workRequestQueue.size();

        if(logDetail.isDebugEnabled()) logDetail.debug("End getRequestQueueSize("+res+")");
        
        return res;
    }
}
