/*
 Copyright (C) 2004 David Bucciarelli (davibu@interfree.it)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.homedns.dade.jcgrid.server;

import java.io.*;
import java.net.*;
import java.util.*;

import org.apache.log4j.*;

import org.homedns.dade.jcgrid.*;
import org.homedns.dade.jcgrid.message.*;
import org.homedns.dade.jcgrid.util.*;

abstract public class HandlerThread extends Thread {
    protected final static String className=HandlerThread.class.getName();
    protected static Logger log=Logger.getLogger(className);
    protected static Logger logDetail=Logger.getLogger("DETAIL."+className);
    
    protected String handlerType;
    
    protected GridServer gridServer;
    protected Socket handlerSocket;
    protected GridMessageChannel handlerChannel;

    protected String sessionName;

    public HandlerThread(String type, GridServer server, Socket socket) throws IOException {
        if(log.isDebugEnabled()) log.debug("Start HandlerThread("+type+","+server+","+socket+")");

        handlerType=type;
        gridServer=server;
        handlerSocket=socket;

        if(gridServer.getNodeConfig().getGridConfig().getUseCompression())
            handlerChannel=new GridMessageGZIPChannel(socket);
        else
            handlerChannel=new GridMessageFlatChannel(socket,true);

        sessionName=null;

        if(log.isDebugEnabled()) log.debug("End HandlerThread()");
    }

    public String getSessionName() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getSessionName()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getSessionName("+sessionName+")");
        
        return sessionName;
    }

    public void run() {
        if(log.isDebugEnabled()) log.debug("Start run()");

        NDC.push("["+handlerType+"]");
        
        try {
            // Waiting for login message

            GridMessageBeginSession gmb=(GridMessageBeginSession)handlerChannel.recv();

            sessionName=gmb.getSessionName();

            NDC.pop();
            NDC.push("["+handlerType+"-"+sessionName+"]");

            String pwd=gridServer.getNodeConfig().getGridConfig().getServerPassword();
            if((pwd!=null) &&
                (!pwd.equals(gmb.getServerPassword())))
                throw new Exception("Login failed for "+sessionName);

            log.warn("New session: "+sessionName+" ["+
                handlerSocket.getLocalAddress().getHostAddress()+":"+
                handlerSocket.getLocalPort()+"]");

            // Check if the session name is valid 
            // If you change the regexp check also the client side code

            if(!sessionName.matches("[a-zA-Z0-9_]+"))
                throw new Exception("Invalid session name: "+sessionName);

            // Check if the session already exists

            HandlerThreads hts;
            if(handlerType.equals(GridNodeConfig.TYPE_CLIENT))
                hts=gridServer.getClientHandlers();
            else if(handlerType.equals(GridNodeConfig.TYPE_WORKER))
                hts=gridServer.getWorkerHandlers();
            else if(handlerType.equals(GridNodeConfig.TYPE_ADMIN))
                hts=gridServer.getAdminHandlers();
            else
                throw new Exception("Unknown handler type: "+handlerType);
            
            if(hts.doesHandlerSessionExists(sessionName,this))
                throw new Exception("Session name already in use: "+sessionName);
            
            // Add to the list of connected clients

            hts.addConnectedHandler(this);
            
            try {
                this.handleConnection();
            } finally {
                log.warn("End session: "+sessionName);
                hts.removeConnectedHandler(this);
            }
        } catch(SocketException sex) {
            if(log.isDebugEnabled()) log.debug("  Socket Exception",sex);
        } catch(EOFException eofex) {
            // Client disconnected

            if(log.isDebugEnabled()) log.debug("  Socket EOF");
        } catch(Exception ex) {
            log.warn("Error in HandlerThread.run()",ex);
        } finally {
            try {
                this.handleDisconnection();
            } catch(Exception ex) {
                log.warn("Failed disconnection of session "+this.getSessionName(),ex);
            }

            try {
                handlerChannel.close();
            } catch(Exception ex) { }
        }

        NDC.pop();
        
        if(log.isDebugEnabled()) log.debug("End run()");
    }
    
    protected void handleConnection() throws Exception {
        // Wait for something to do

        for(;;) {
            GridMessage msg=handlerChannel.recv();

            if(log.isDebugEnabled()) log.debug("  Received grid message: "+msg);

            if(msg instanceof GridMessagePing) {
                if(log.isDebugEnabled()) log.debug("  Received ping request");

                handlerChannel.send(new GridMessagePingAck());
            } else if(msg instanceof GridMessageEndSession) {
                if(log.isDebugEnabled()) log.debug("  Received end session request");
                break;
            } else
                this.handleMsg(msg);
        }
    }
    
    protected void handleMsg(GridMessage msg) throws Exception {
        log.warn("Unknown message in HandlerThread.handleMsg(): "+msg);
    }

    protected void handleDisconnection() throws Exception {
        // Do nothig by default
    }
}
