/*
 Copyright (C) 2004 David Bucciarelli (davibu@interfree.it)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.homedns.dade.jcgrid.util;

import java.io.*;
import java.net.*;
import java.util.zip.*;

import org.apache.log4j.*;

import org.homedns.dade.jcgrid.message.*;

public class GridMessageGZIPChannel implements GridMessageChannel {
    private final static String className=GridMessageGZIPChannel.class.getName();
    private static Logger log=Logger.getLogger(className);
    private static Logger logDetail=Logger.getLogger("DETAIL."+className);

    private Socket socket;
    private InputStream is;
    private OutputStream os;

    public GridMessageGZIPChannel(Socket s) throws IOException {
        if(log.isDebugEnabled()) log.debug("Start GridMessageGZIPChannel("+s+")");
        
        socket=s;

        is=socket.getInputStream();
        os=socket.getOutputStream();
        
        if(log.isDebugEnabled()) log.debug("End GridMessageGZIPChannel()");
    }
    
    private int readPacket(byte[] buf) throws Exception {
        if(log.isDebugEnabled()) log.debug("Start readPacket("+buf+")");
        
        int base=0;
        while(buf.length-base>0) {
            int res=is.read(buf,base,buf.length-base);
            
            if(res==-1) {
                if(log.isDebugEnabled()) log.debug("End readPacket(-1)");

                return -1;
            }
            
            base+=res;
        }

        if(log.isDebugEnabled()) log.debug("End readPacket("+buf.length+")");

        return buf.length;
    }
    
    public GridMessage recv() throws Exception {
        if(log.isDebugEnabled()) log.debug("Start recv()");

        byte[] bufSize=new byte[4];
        if(this.readPacket(bufSize)!=4)
            throw new Exception("Error reading a message header (size field)");
        int size=(bufSize[0] & 0xff) |
            ((bufSize[1] & 0xff)<<8) |
            ((bufSize[2] & 0xff)<<16) |
            ((bufSize[3] & 0xff)<<24);
        if(log.isDebugEnabled()) log.debug("  Packet size: "+size);

        byte[] bufData=new byte[size];
        if(this.readPacket(bufData)!=size)
            throw new Exception("Error reading a message (payload field)");

        ByteArrayInputStream bis=new ByteArrayInputStream(bufData);
        GZIPInputStream gis=new GZIPInputStream(bis);
        ObjectInputStream ois=new ObjectInputStream(gis);
        GridMessage msg=(GridMessage)ois.readObject();

        if(log.isDebugEnabled()) log.debug("End recv("+msg+")");
        
        return msg;
    }

    public GridMessage recv(int timeout) throws Exception {
        if(log.isDebugEnabled()) log.debug("Start recv("+timeout+")");

        socket.setSoTimeout(timeout);
        GridMessage msg=this.recv();
        socket.setSoTimeout(0);

        if(log.isDebugEnabled()) log.debug("End recv("+msg+")");

        return msg;
    }

    public void send(GridMessage msg) throws Exception  {
        if(log.isDebugEnabled()) log.debug("Start send("+msg+")");

        ByteArrayOutputStream bos=new ByteArrayOutputStream();
        ObjectOutputStream oos=new ObjectOutputStream(bos);
        oos.writeObject(msg);
        oos.close();
        byte[] obj=bos.toByteArray();
        
        if(log.isDebugEnabled()) log.debug("  Array size: "+obj.length);

        // Compress data

        bos=new ByteArrayOutputStream();
        GZIPOutputStream gos=new GZIPOutputStream(bos);
        gos.write(obj);
        gos.close();
        obj=bos.toByteArray();

        if(log.isDebugEnabled()) log.debug("  Array compressed size: "+obj.length);

        // Send header (size) and data

        byte[] buf=new byte[4];
        buf[0]=(byte)(obj.length & 0x000000ff);
        buf[1]=(byte)((obj.length & 0x0000ff00)>>8);
        buf[2]=(byte)((obj.length & 0x00ff0000)>>16);
        buf[3]=(byte)((obj.length & 0xff000000)>>24);
        os.write(buf);

        os.write(obj);
        os.flush();

        if(log.isDebugEnabled()) log.debug("End send()");
    }
    
    public void close() throws IOException {
        if(log.isDebugEnabled()) log.debug("Start close()");

        is.close();
        os.close();
        socket.close();

        if(log.isDebugEnabled()) log.debug("End close()");
    }
}
