/*
 Copyright (C) 2004 David Bucciarelli (davibu@interfree.it)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.homedns.dade.jcgrid.vfs;

import java.io.*;
import java.util.*;
import java.security.*;

import org.apache.log4j.*;

import org.homedns.dade.jcgrid.message.*;
import org.homedns.dade.jcgrid.util.*;

public class vfsSession implements Serializable {
    private final static String className=vfsSession.class.getName();
    private static Logger log=Logger.getLogger(className);
    private static Logger logDetail=Logger.getLogger("DETAIL."+className);

    private static final long serialVersionUID=1L;
    
    private final static String hexMap[]={
        "0","1","2","3","4","5","6","7","8","9",
        "A","B","C","D","E","F"
    };
    
    private String path;
    private Hashtable fileList;
    private long size;
    private boolean inUse;

    public vfsSession(String dir) throws Exception {
        if(log.isDebugEnabled()) log.debug("Start vfsSession("+dir+")");

        path=dir;
        
        // Check if the directory exists
        
        File pathDir=new File(path);
        if(!pathDir.exists()) {
            // I fit doesn't exist create the directory
            pathDir.mkdirs();
        }
        
        fileList=null;
        inUse=false;

        if(log.isDebugEnabled()) log.debug("End vfsSession()");
    }

    //---------------------------- Delete --------------------------------------

    private void deleteImpl(File root) throws IOException {
        if(log.isDebugEnabled()) log.debug("Start deleteImpl("+root+")");

        String[] l=root.list();
        for(int i=0;i<l.length;i++) {
            String p=root.getAbsolutePath()+File.separator+l[i];

            File f=new File(p);
            if(f.isDirectory()) {
                log.warn("VFS session removing directory: "+p);
                this.deleteImpl(f);
                f.delete();
            } else
                f.delete();
        }

        boolean res=(l.length==0);
        
        if(log.isDebugEnabled()) log.debug("End deleteImpl()");
    }
    
    public void delete() throws IOException {
        if(log.isDebugEnabled()) log.debug("Start delete()");

        File f=new File(path);
        this.deleteImpl(f);

        if(log.isDebugEnabled()) log.debug("End delete()");

    }

    private boolean deleteEmptyDirectoriesImpl(File root) throws IOException {
        if(log.isDebugEnabled()) log.debug("Start deleteEmptyDirectoriesImpl("+root+")");

        String[] l=root.list();
        for(int i=0;i<l.length;i++) {
            String p=root.getAbsolutePath()+File.separator+l[i];

            File f=new File(p);
            if(f.isDirectory()) {
                if(this.deleteEmptyDirectoriesImpl(f)) {
                    log.warn("VFS session removing empty directory: "+p);
                    f.delete();
                }
            }
        }

        boolean res=(l.length==0);
        
        if(log.isDebugEnabled()) log.debug("End deleteEmptyDirectoriesImpl("+res+")");

        return res;
    }
    
    public void deleteEmptyDirectories() throws IOException {
        if(log.isDebugEnabled()) log.debug("Start deleteEmptyDirectories()");

        File f=new File(path);
        this.deleteEmptyDirectoriesImpl(f);

        if(log.isDebugEnabled()) log.debug("End deleteEmptyDirectories()");

    }

    //--------------------------------------------------------------------------

    public void removeNode(String name)  {
        if(log.isDebugEnabled()) log.debug("Start removeNode("+name+")");

        vfsNode n=(vfsNode)fileList.get(name);
        if(n!=null) {
            size-=n.getSize();
            fileList.remove(name);
        }

        File f=new File(path+name);
        f.delete();

        if(log.isDebugEnabled()) log.debug("End removeNode()");

    }

    private String toHexString(byte[] data) {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start toHexString("+data+")");

        StringBuffer result=new StringBuffer();
            
        for(int i=0;i<data.length;i++) {
            int l=((((int)data[i])) & 0xF0)>>4;
            int r=((int)data[i]) & 0x0F;
            
            result.append(hexMap[l]+hexMap[r]);
        }

        String res=result.toString();

        if(logDetail.isDebugEnabled()) logDetail.debug("End toHexString("+res+")");

        return res;
    }

    private String generateUniqueID(File f) throws IOException,
        NoSuchAlgorithmException, NoSuchProviderException {
        if(log.isDebugEnabled()) log.debug("Start generateUniqueID("+f+")");

        byte[] buf=new byte[8192];
        int bytesRead;

        FileInputStream fis=new FileInputStream(f);
        MessageDigest md5=MessageDigest.getInstance("MD5","SUN");

        while((bytesRead=fis.read(buf))!=-1) {
            md5.update(buf,0,bytesRead);
        }

        fis.close();
        String res=this.toHexString(md5.digest());

        if(log.isDebugEnabled()) log.debug("End generateUniqueID("+res+")");

        return res;
    }

    public void addNode(String name) throws IOException,
        NoSuchAlgorithmException, NoSuchProviderException {
        if(log.isDebugEnabled()) log.debug("Start addNode("+name+")");

        File f=new File(path+name);        
        if(log.isDebugEnabled()) log.debug("  Name: "+name);

        String id=this.generateUniqueID(new File(path+name));
        if(log.isDebugEnabled()) log.debug("  MD5: "+id);

        long s=f.length();
        if(log.isDebugEnabled()) log.debug("  Size: "+s);

        fileList.put(name,new vfsNode(name,id,s));
        size+=s;

        if(log.isDebugEnabled()) log.debug("End addNode()");

    }

    public Vector notIncludedNodes(vfsSession s)  {
        if(log.isDebugEnabled()) log.debug("Start notIncludedNodes("+s+")");

        Vector v=new Vector();
        for(Iterator i=fileList.keySet().iterator();i.hasNext();) {
            String name=(String)i.next();
            vfsNode n=(vfsNode)fileList.get(name);
            
            if(!s.includes(n))
                v.add(name);
        }

        if(log.isDebugEnabled()) log.debug("End notIncludedNodes("+v+")");

        return v;
    }

    public boolean includes(vfsNode node)  {
        if(log.isDebugEnabled()) log.debug("Start includes("+node+")");

        boolean res;

        vfsNode n=(vfsNode)fileList.get(node.getName());
        if(n==null)
            res=false;
        else
            res=n.equals(node);

        if(log.isDebugEnabled()) log.debug("End includes("+res+")");

        return res;
    }

    public void syncVFSSession(GridMessageChannel serverChannel) throws Exception {
        if(log.isDebugEnabled()) log.debug("Start syncVFSSession("+serverChannel+")");

        // Sending the list of files
        
        serverChannel.send(new GridMessageVFSSessionBegin(this));

        // Wait for file requests
        
        for(;;) {
            GridMessage msg=serverChannel.recv();

            if(log.isDebugEnabled()) log.debug("  Received grid message: "+msg);
            
            if(msg instanceof GridMessageVFSSessionEnd) {
                if(log.isDebugEnabled()) log.debug("  Received VFS session end");
                break;
            } else if(msg instanceof GridMessageVFSSessionFileRequest) {
                if(log.isDebugEnabled()) log.debug("  Received VFS session file request");

                String n=((GridMessageVFSSessionFileRequest)msg).getName();
                log.warn("VFS session sending: "+n);

                // Read the file

                File f=new File(path+n);
                long fsize=f.length();
                if(log.isDebugEnabled()) log.debug("  File size: "+fsize);
                // TODO: 4GB limit
                byte[] data=new byte[(int)fsize];

                FileInputStream fis=new FileInputStream(f);
                fis.read(data);
                fis.close();

                serverChannel.send(new GridMessageVFSSessionFileResult(data));
            } else
                log.warn("Unknown message in ClientHandlerThread.run(): "+msg);
        }

        if(log.isDebugEnabled()) log.debug("End syncVFSSession()");
    }
    
    //----------------------------- Start & Stop -------------------------------

    private void scann(File root, int realRootPathIndex) throws IOException,
        NoSuchAlgorithmException, NoSuchProviderException {
        if(log.isDebugEnabled()) log.debug("Start scann("+root+","+realRootPathIndex+")");

        String[] l=root.list();

        for(int i=0;i<l.length;i++) {
            String p=root.getAbsolutePath()+File.separator+l[i];
            log.warn("  VFS scanning: "+p);

            File f=new File(p);
            if(f.isDirectory())
                this.scann(f,realRootPathIndex);
            else if(f.isFile()) {
                String name=p.substring(realRootPathIndex);
                this.addNode(name);
            } else
                log.warn("    VFS "+l[i]+" is not a normal file");
        }

        if(log.isDebugEnabled()) log.debug("End scann()");
    }
    
    public void start() throws Exception {
        if(log.isDebugEnabled()) log.debug("Start start()");

        fileList=new Hashtable();
        size=0;

        log.warn("VFS starting on path: "+path);
        File f=new File(path);
        
        if(!f.exists())
            throw new Exception("The path "+path+" doesn't exist");
        
        this.scann(f,f.getAbsolutePath().length());
        
        if(log.isDebugEnabled()) log.debug("End start()");
    }

    //------------------------------- Get & Set --------------------------------

    public vfsNode getNode(String name) {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getNode("+name+")");

        vfsNode n=(vfsNode)fileList.get(name);

        if(logDetail.isDebugEnabled()) logDetail.debug("End getNode("+n+")");
        
        return n;
    }

    public String getPath() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getPath()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getPath("+path+")");
        
        return path;
    }

    public long getSize() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getSize()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getSize("+size+")");
        
        return size;
    }

    public void setInUse(boolean use) {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start setInUse("+use+")");

        inUse=use;

        if(logDetail.isDebugEnabled()) logDetail.debug("End setInUse()");
    }

    public boolean getInUse() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getInUse()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getInUse("+inUse+")");
        
        return inUse;
    }
}
