/*
 Copyright (C) 2004 David Bucciarelli (davibu@interfree.it)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.homedns.dade.jcgrid.vfs;

import java.io.*;
import java.util.*;
import java.security.*;

import org.apache.log4j.*;

import org.homedns.dade.jcgrid.message.*;
import org.homedns.dade.jcgrid.util.*;

public class vfsSessionPool {
    private final static String className=vfsSessionPool.class.getName();
    private static Logger log=Logger.getLogger(className);
    private static Logger logDetail=Logger.getLogger("DETAIL."+className);

    private String path;
    private Hashtable sessionList;
    private long size;
    private long maxSize;

    public vfsSessionPool(String dir, long s) {
        if(log.isDebugEnabled()) log.debug("Start vfsSessionPool("+dir+","+s+")");

        path=dir;

        // Check if the directory exists
        
        File pathDir=new File(path);
        if(!pathDir.exists()) {
            // I fit doesn't exist create the directory
            pathDir.mkdirs();
        }
        
        sessionList=null;
        maxSize=s;

        if(log.isDebugEnabled()) log.debug("End vfsSessionPool()");
    }

    //--------------------------------------------------------------------------

    // Very simple but it works
    private boolean garbageCollect() throws IOException {
        if(log.isDebugEnabled()) log.debug("Start garbageCollect()");

        boolean res=false;
        synchronized(sessionList) {
            for(Iterator i=sessionList.keySet().iterator();i.hasNext();) {
                String name=(String)i.next();
                vfsSession s=(vfsSession)sessionList.get(name);
                
                if(!s.getInUse()) {
                    // Remove this session

                    log.warn("VFS removing unused session: "+name);
                    
                    size-=s.getSize();
                    s.delete();
                    sessionList.remove(name);
                    String dir=path+File.separator+name;
                    File f=new File(dir);
                    f.delete();

                    res=true;
                    break;
                }
            }
        }

        if(log.isDebugEnabled()) log.debug("End garbageCollect("+res+")");

        return res;
    }
    
    public void syncVFSSession(String name, GridMessageChannel clientChannel,
        vfsSession clientSession) throws Exception {
        if(log.isDebugEnabled()) log.debug("Start syncVFSSession("+name+","+clientChannel+","+clientSession+")");

        if(log.isDebugEnabled()) log.debug("  Cache size: "+size/(1024*1024)+"/"+maxSize/(1024*1024)+"MB");

        // Reject session that will never fit in the cache

        if(maxSize<clientSession.getSize())
            throw new Exception(
                "Cache too small: "+
                maxSize+"<"+
                clientSession.getSize());

        // Check if there is enough space in the cache for the new
        // session (worst case: all file are new)

        if(maxSize<size+clientSession.getSize()) {
            // Try to make some free space

            while(this.garbageCollect()) {
                if(log.isDebugEnabled()) log.debug("  Cache size: "+size);

                if(maxSize>size+clientSession.getSize())
                    break;
            }

            if(maxSize<size+clientSession.getSize())
                throw new Exception(
                    "Not enough cache free space: "+
                    maxSize+"<"+
                    size+"+"+
                    clientSession.getSize());
        }

        if(log.isDebugEnabled()) log.debug("  Cache size: "+size/(1024*1024)+"/"+maxSize/(1024*1024)+"MB");

        vfsSession cacheSession=this.lockVFSSession(name);
        if(cacheSession==null)
            cacheSession=this.newVFSSession(name);

        // Looks for file to delete

        Vector toDelete=cacheSession.notIncludedNodes(clientSession);
        for(Iterator i=toDelete.iterator();i.hasNext();) {
            String nodeName=(String)i.next();
            String fileName=path+File.separator+name+nodeName;
            log.warn("VFS deleting: "+fileName);

            synchronized(sessionList) {
                size-=cacheSession.getNode(nodeName).getSize();
            }
            cacheSession.removeNode(nodeName);
        }

        // Looks for empty directories

        cacheSession.deleteEmptyDirectories();

        // Looks for new files

        Vector toUpdate=clientSession.notIncludedNodes(cacheSession);
        for(Iterator i=toUpdate.iterator();i.hasNext();) {
            String nodeName=(String)i.next();
            String fileName=path+File.separator+name+nodeName;
            log.warn("VFS updating: "+fileName);

            // Request file
            
            clientChannel.send(new GridMessageVFSSessionFileRequest(nodeName));

            // Receive file

            GridMessageVFSSessionFileResult gmf=(GridMessageVFSSessionFileResult)
                clientChannel.recv();

            // Create the directories

            File f=new File(fileName);
            String fparent=f.getParent();
            if(fparent!=null) {
                f=new File(fparent);
                f.mkdirs();
            }

            // Write file

            FileOutputStream fos=new FileOutputStream(new File(fileName));
            fos.write(gmf.getData());
            fos.close();

            // Update session

            cacheSession.addNode(nodeName);
            synchronized(sessionList) {
                size+=cacheSession.getNode(nodeName).getSize();
            }
        }

        log.warn("VFS session cache size: "+size/(1024*1024)+"/"+maxSize/(1024*1024)+"MB");

        if(log.isDebugEnabled()) log.debug("End syncVFSSession()");
    }
    
    public void start() throws Exception {
        if(log.isDebugEnabled()) log.debug("Start start()");

        sessionList=new Hashtable();

        log.warn("VFS session pool on path: "+path);
        File f=new File(path);
        String[] l=f.list();

        for(int i=0;i<l.length;i++) {
            String p=f.getAbsolutePath()+File.separator+l[i];
            log.warn("VFS session pool scanning: "+p);

            File s=new File(p);
            if(s.isDirectory()) {
                vfsSession vfss=new vfsSession(p);
                vfss.start();

                String name=l[i];
                sessionList.put(name,vfss);
                size+=vfss.getSize();
            } else
                log.warn("    VFS "+l[i]+" is not a directory");
        }
        
        if(log.isDebugEnabled()) log.debug("End start()");
    }

    public vfsSession newVFSSession(String name) throws Exception {
        if(log.isDebugEnabled()) log.debug("Start newVFSSesssion("+name+")");
        
        String dir=path+File.separator+name;
        File f=new File(dir);
        f.mkdir();

        vfsSession s=new vfsSession(dir);
        s.start();
        s.setInUse(true);

        synchronized(sessionList) {
            sessionList.put(name,s);
            size+=s.getSize();
        }
        
        if(log.isDebugEnabled()) log.debug("End newVFSSesssion("+s+")");
        
        return s;
    }

    //----------------------------- Lock & Unlock ------------------------------

    public vfsSession lockVFSSession(String name) {
        if(log.isDebugEnabled()) log.debug("Start lockVFSSession("+name+")");
        
        vfsSession s;
        synchronized(sessionList) {
            s=(vfsSession)sessionList.get(name);
            if(s!=null)
                s.setInUse(true);
        }
        
        if(log.isDebugEnabled()) log.debug("End lockVFSSession("+s+")");
        
        return s;
    }

    public void unlockVFSSession(String name) {
        if(log.isDebugEnabled()) log.debug("Start unlockVFSSession("+name+")");

        synchronized(sessionList) {
            vfsSession s=(vfsSession)sessionList.get(name);
            if(s!=null)
                s.setInUse(false);
        }

        if(log.isDebugEnabled()) log.debug("End unlockVFSSession()");
    }

    //------------------------------- Get & Set --------------------------------

    public vfsSession getVFSSession(String name) {
        if(log.isDebugEnabled()) log.debug("Start getVFSSession("+name+")");
        
        vfsSession s=(vfsSession)sessionList.get(name);
        
        if(log.isDebugEnabled()) log.debug("End getVFSSession("+s+")");
        
        return s;
    }

    public String getPath() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getPath()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getPath("+path+")");
        
        return path;
    }

    public long getSize() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getSize()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getSize("+size+")");
        
        return size;
    }

    public long getMaxSize() {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start getMaxSize()");
        if(logDetail.isDebugEnabled()) logDetail.debug("End getMaxSize("+maxSize+")");
        
        return maxSize;
    }
    
    public void setMaxSize(long s) {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start setMaxSize()");

        maxSize=s;

        if(logDetail.isDebugEnabled()) logDetail.debug("End setMaxSize()");
    }
}
