/*
 Copyright (C) 2004 David Bucciarelli (davibu@interfree.it)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.homedns.dade.jcgrid.worker;

import java.io.*;
import java.net.*;
import java.util.*;

import org.apache.log4j.*;

import org.homedns.dade.jcgrid.*;
import org.homedns.dade.jcgrid.vfs.*;
import org.homedns.dade.jcgrid.message.*;
import org.homedns.dade.jcgrid.util.*;

public class WorkerThread extends Thread {
    private final static String className=WorkerThread.class.getName();
    private static Logger log=Logger.getLogger(className);
    private static Logger logDetail=Logger.getLogger("DETAIL."+className);

    private GridWorker gridWorker;
    private GridWorkerFeedback feedback;

    public WorkerThread(GridWorker gw) throws IOException {
        if(log.isDebugEnabled()) log.debug("Start WorkerThread("+gw+")");

        gridWorker=gw;
        feedback=null;

        if(log.isDebugEnabled()) log.debug("End WorkerThread()");
    }

    public void setWorkerFeedback(GridWorkerFeedback f) {
        if(logDetail.isDebugEnabled()) logDetail.debug("Start setWorkerFeedback("+f+")");

        feedback=f;

        if(logDetail.isDebugEnabled()) logDetail.debug("End setWorkerFeedback()");
    }

    public void run() {
        if(log.isDebugEnabled()) log.debug("Start run()");

        try {
            if(feedback!=null)
                feedback.start();

            GridMessageChannel serverChannel=gridWorker.getGridMessageChannel();

            for(;;) {
                GridMessage msg=serverChannel.recv();

                try {
                    if(msg instanceof GridMessageShutdown)
                        break;
                    else if(msg instanceof GridMessagePing)
                        serverChannel.send(new GridMessagePingAck());
                    else if(msg instanceof GridMessageWorkRequest) {
                        GridMessageWorkRequest gmwr=(GridMessageWorkRequest)msg;

                        WorkRequest wr=gmwr.getWorkRequest();
                        if(log.isDebugEnabled()) log.debug("  Received work request: "+wr);

                        String sessionName=wr.getSessionName();
                        log.warn("Working for: "+sessionName);

                        if(feedback!=null)
                            feedback.beginWorkingFor(sessionName,wr);
                        
                        String sessionPath=null;
                        if(gridWorker.getNodeConfig().getGridConfig().getUseVFS()) {
                            // Wait for VFS Session

                            GridMessage vfsmsg=serverChannel.recv();

                            if(log.isDebugEnabled()) log.debug("  Received VFS session begin");

                            vfsSession clientSession=((GridMessageVFSSessionBegin)vfsmsg).getVFSSession();

                            // Align sessions

                            gridWorker.getVFSSessionPool().syncVFSSession(
                                sessionName,
                                serverChannel,
                                clientSession);

                            // End file exchange

                            serverChannel.send(new GridMessageVFSSessionEnd());

                            sessionPath=gridWorker.getVFSSessionPool().
                                getVFSSession(sessionName).getPath();
                        }

                        // Do the work

                        WorkResult res=gridWorker.getWorker().doWork(wr,sessionPath);
                        if(log.isDebugEnabled()) log.debug("  Work result: "+res);

                        if(feedback!=null)
                            feedback.endWorkingFor(res);
                        
                        // Return the result

                        serverChannel.send(
                            new GridMessageWorkResult(res));
                    } else
                        log.warn("Error in WorkerThread.run(), received an unknown message: "+msg);
                } catch(Exception ex) {
                    log.warn("Error in WorkerThread.run.loop()",ex);
                }
            }
        } catch(Exception ex) {
            log.warn("Error in WorkerThread.run()",ex);
        }

        if(log.isDebugEnabled()) log.debug("End run()");
    }
}
