/*
 Copyright (C) 2004 David Bucciarelli (davibu@interfree.it)

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

package org.homedns.dade.jcgrid.worker.impl.povray;

import java.util.*;

import org.homedns.dade.jcgrid.message.*;
import org.homedns.dade.jcgrid.client.*;

public class POVRenderingClient extends Thread {
    private GridNodeClientConfig clientCfg;
    private POVRenderingClientFeedback clientFeedback;

    private POVWorkRequest povWReq;
    private int fragSize;
    private boolean hasAnimation;
    private int startFrame;
    private int endFrame;

    private POVRenderingFrame[] frames;
    
    public POVRenderingClient(
        GridNodeClientConfig cfg,
        POVRenderingClientFeedback feedback,
        POVWorkRequest req,
        int frag) {
        clientCfg=cfg;
        clientFeedback=feedback;

        povWReq=req;
        fragSize=frag;
        hasAnimation=false;
        
        frames=new POVRenderingFrame[1];
        frames[0]=null;
    }

    public POVRenderingClient(
        GridNodeClientConfig cfg,
        POVRenderingClientFeedback feedback,
        POVWorkRequest req,
        int frag, int start, int end) {
        this(cfg,feedback,req,frag);
        
        hasAnimation=true;
        startFrame=start;
        endFrame=end;

        frames=new POVRenderingFrame[endFrame-startFrame+1];
        Arrays.fill(frames,null);
    }

    public void run() {
        try {
            // Start Client
            
            GridClient gc=new GridClient();
            gc.setNodeConfig(clientCfg);
            gc.start();

            long t1=System.currentTimeMillis();
            try {
                // Splitting the work
                
                int xStep=fragSize;
                int yStep=(int)(fragSize*povWReq.getHeight()/(double)povWReq.getWidth());
                POVWorkRequest[] workList;
                if(hasAnimation)
                    workList=povWReq.split(startFrame,endFrame,xStep,yStep);
                else
                    workList=povWReq.split(xStep,yStep);

                int fragCount=povWReq.fragmentPerFrame(xStep,yStep);

                clientFeedback.setProgressMaximum(0);
                clientFeedback.setProgressMaximum(2*workList.length-1);
                
                clientFeedback.renderingBegin();
                
                // Sending work requests

                for(int i=0;i<workList.length;i++) {
                    clientFeedback.setProgressValue(i);
                    clientFeedback.sendingFrgamentRequest(workList[i]);

                    gc.send(new GridMessageWorkRequest(workList[i]));

                    if(this.isInterrupted())
                        break;
                }

                // Receiving work results

                for(int i=0;i<workList.length;i++) {
                    clientFeedback.setProgressValue(i+workList.length);

                    GridMessageWorkResult gmwr=(GridMessageWorkResult)gc.recv();
                    POVWorkResult povWRes=(POVWorkResult)gmwr.getWorkResult();

                    int idx=povWRes.getRID()-1;
                    int frameIdx=workList[idx].getFrame()-startFrame;

                    if(frames[frameIdx]==null) {
                        // If the framebuffer doesn't exist create it

                        frames[frameIdx]=new POVRenderingFrame(
                            workList[idx].getWidth(),workList[idx].getHeight(),fragCount);
                    }
                    frames[frameIdx].addFragmentResult(workList[idx],povWRes);

                    clientFeedback.receivedFragmentResult(workList[idx],povWRes,
                        workList[idx].getFrame(),frames[frameIdx]);

                    if(frames[frameIdx].isComplete()) {
                        clientFeedback.completeFrame(workList[idx].getFrame(),frames[frameIdx]);

                        // Free memory

                        frames[frameIdx]=null;
                    }

                    if(this.isInterrupted())
                        break;
                }
            } finally {
                long t2=System.currentTimeMillis();

                clientFeedback.setRenderingTime(povWReq,t2-t1);

                try {
                    gc.stop();
                } catch(Exception ex) { }
            }
        } catch(Exception ex) {
            clientFeedback.error("Error while doing the rendering",ex);
        }

        clientFeedback.renderingEnd();
    }
}
